package Test2::Harness;
use strict;
use warnings;

our $VERSION = '0.999006'; # TRIAL

1;

__END__

=pod

=encoding UTF-8

=head1 NAME

Test2::Harness - A new and improved test harness with better L<Test2>
integration.

=head1 DESCRIPTION

Test2::Harness is the backend code that handles running/processing the tests.
In general a user will not use it directly, instead you should probably be
looking at L<App::Yath> which is the UI layer built around Test2::Harness.

=head1 SEE ALSO

The primary documentation can be found in L<App::Yath>.

=head1 SOURCE

The source code repository for Test2-Harness can be found at
F<http://github.com/Test-More/Test2-Harness/>.

=head1 MAINTAINERS

=over 4

=item Chad Granum E<lt>exodist@cpan.orgE<gt>

=back

=head1 AUTHORS

=over 4

=item Chad Granum E<lt>exodist@cpan.orgE<gt>

=back

=head1 COPYRIGHT

Copyright 2020 Chad Granum E<lt>exodist7@gmail.comE<gt>.

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

See F<http://dev.perl.org/licenses/>

=cut
